/*
 * Copyright (c) 1998 by Tech Soft 3D, LLC.
 * The information contained herein is confidential and proprietary to
 * Tech Soft 3D, LLC., and considered a trade secret as defined under
 * civil and criminal statutes.  Tech Soft 3D shall pursue its civil
 * and criminal remedies in the event of unauthorized use or misappropriation
 * of its trade secrets.  Use of this information by anyone other than
 * authorized employees of Tech Soft 3D, LLC. is granted only under a
 * written non-disclosure agreement, expressly prescribing the scope and
 * manner of such use.
 *
 * $Id$
 */

/* 
 * This program is a demo for the new function, compute_optimized_shell  
 */

#include "hc.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


/*
 *	This program reads a HOOPS metafile and searches it
 *	for polygons.  Polygons found are inserted into a
 *	Shell and then the Shell is optimized. The resulting
 *	optimized Shell is written to a new HOOPS metafile.
 *
 *	The program takes two command-line arguments
 *	in-file and out-file
 */

typedef struct point {
	float	x, y, z;
} Point;

/* it just so happens that point and vector are exactly alike */
typedef struct point Vector; 

typedef struct SHELL_INFO_TAG {
	int point_list_len;
	Point *point_list;
	Vector *normal_list;
	int face_list_len;
	int *face_list;
} Shell_Info;

#define ISRECURSIVE					1
#define MAX_SEGMENT_NAME_LENGTH		1024
#define MAX_TYPE_NAME_LENGTH		32
#define streq(a, b) (strcmp(a, b) == 0)

/* temporary workspace */
char	child[ MAX_SEGMENT_NAME_LENGTH ];

/*******************************\
 * HELPER (recursive)
 * Shell_Search
 * Designed to be called from Segment_Allocate.
 * Finds how much memory necessary for the current and child segments' shells.
 * Fills two structures:
 *   total_info is designed to be big enough for all of the shells put together
 *   max_info should be big enough for the largest shell.
 * Note that only the lengths are valid after this call, not the actual lists.
\*******************************/
void Shell_Search( Shell_Info *total_info, Shell_Info *max_info, int do_recurse )
{
	int		local_point_list_len;
	int		local_face_list_len;
	HC_KEY		key;
	int		num_shells = 0;
	char	type[ MAX_TYPE_NAME_LENGTH ];
	
	HC_Begin_Segment_Search ("*");
	if (do_recurse)	
	{
	    while (HC_Find_Segment (child) ) 
		{
			HC_Open_Segment (child);
			Shell_Search (total_info, max_info, do_recurse);
			HC_Close_Segment();
		}
	}
	HC_End_Segment_Search();
	
	HC_Begin_Contents_Search (".", "shells");
	HC_Show_Contents_Count (&num_shells);
	while (HC_Find_Contents (type, &key)) 
	{
		HC_Show_Shell_Size(key, &local_point_list_len, &local_face_list_len );
		/* keep track of the totals  */
		total_info -> point_list_len += local_point_list_len;
		total_info -> face_list_len  += local_face_list_len;
		/* keep track of the maximums */
		if( max_info -> point_list_len < local_point_list_len )
			max_info -> point_list_len = local_point_list_len;
		if( max_info -> face_list_len < local_face_list_len )
			max_info -> face_list_len = local_face_list_len;
	}
	HC_End_Contents_Search ();
	
} /* end function Shell_Search */


/*******************************\
 * HELPER
 * Segment_Shell_Allocate
 * allocates memory for the current and child segments' shells.
 * Allocates two structures:
 *   total_info is designed to be big enough for all of the shells put together
 *   max_info should be big enough for the largest shell.
 * After the memory is allocated, the length fields are cleared.
 * returns 0 if the memory could not be allocated
 \*******************************/
int Segment_Shell_Allocate( Shell_Info *total_info, Shell_Info *max_info, int do_recurse )
{
	Shell_Search (total_info, max_info, do_recurse);
	
	max_info -> point_list = (Point *) 
		malloc ( max_info -> point_list_len * sizeof (Point) );
	max_info -> normal_list = (Vector *)
		malloc ( max_info -> point_list_len * sizeof (Vector) );
	max_info -> face_list  = (int *)   
		malloc ( max_info -> face_list_len * sizeof (int) );
	total_info -> point_list = (Point *) 
		malloc ( total_info -> point_list_len * sizeof (Point) );
	total_info -> normal_list = (Vector *)
		malloc ( total_info -> point_list_len * sizeof (Vector) );
	total_info -> face_list  = (int *)   
		malloc ( total_info -> face_list_len * (long) sizeof (int) );
	
	/* clear the lengths of the structures */
	max_info -> point_list_len = 0;
	max_info -> face_list_len = 0;
	total_info -> point_list_len = 0;
	total_info -> face_list_len = 0;
	
	/* return 0 if malloc returned NULL for any of the memory requests */
	if(  (max_info -> point_list) &&
	   (max_info -> face_list) &&
	   (total_info -> point_list) &&
	   (total_info -> face_list) )
		return 1;
	else
		return 0;
	
} /* end function Segment_Shell_Allocate */


/*******************************\
 * HELPER (recursive)
 * Recursive_Merge_Shells
 * This function assumes that the proper amount of memory has 
 *      been allocated (see Segment_Shell_Allocate)
 * Recursive_Merge steps through the child segments, and merges all of the
 *      shells into one.
 * XXX this routine leaves a lot of empty segments in their place.  Ideally,
 *      we would like to step through and delete them.  That will be the 
 *      next step.
 \*******************************/
void Recursive_Merge_Shells( Shell_Info *total_info, Shell_Info *max_info, int do_recurse )
{
	int		local_point_list_len;
	int		local_face_list_len;
	int		polygon_length;
	int		i, j;
	char	type[ MAX_TYPE_NAME_LENGTH ];
	HC_KEY	key;
	Vector	*v;
	int		current_point_list_offset, current_face_list_offset;

	/* recursively find all of the child segments */
	HC_Begin_Segment_Search ("*");
    if (do_recurse) 
	{
		while (HC_Find_Segment (child) ) 
		{
			HC_Open_Segment (child);
			Recursive_Merge_Shells (total_info, max_info, do_recurse  );
			HC_Close_Segment ();
		}
	}
	HC_End_Segment_Search ();
	
	/*
	 * after all of the children have been processed, process  current segment
	 *   (i.e., a post-order tree traversal)
	 */
	HC_Begin_Contents_Search (".", "shells");
	while (HC_Find_Contents (type, &key)) 
	{
		current_point_list_offset = total_info -> point_list_len;
		current_face_list_offset  = total_info -> face_list_len;
		
		/*
		 * use max_info for the scratch-space buffers, but
		 *  write the lengths into local variables, just in case 
		 *  we need the max lengths later on.
		 */
		HC_Show_Shell ( key, 
					    &(local_point_list_len), 
					    max_info -> point_list, 
					    &(local_face_list_len), 
					    max_info -> face_list 
					    );

		/* 
		 * Copy the point and face list into the main face list
		 * For each individual shell, the point list originally started 
		 *   at zero, but for the main list, it starts at 
		 *   current_point_list_offset.  
		 */
		/* here, we copy the point list */
		for( i = 0 ; i < local_point_list_len ; i++ )
		{
			total_info -> point_list[ current_point_list_offset + i].x = 
				max_info -> point_list[i].x;
			total_info -> point_list[ current_point_list_offset + i].y = 
				max_info -> point_list[i].y;
			total_info -> point_list[ current_point_list_offset + i].z = 
				max_info -> point_list[i].z;
		}
		
		/* here, we copy the face list */
		i = 0;
		while(i < local_face_list_len)
		{
			/*
			 * in this loop:
			 * main_face_list[current_face_list_offset]  
			 *      is the start of the individual shell
			 * main_face_list[current_face_list_offset + i]  
			 *     is the start of a polygon inside the shell
			 * main_face_list[current_face_list_offset + i + j]
			 *      is a point inside the polygon inside the shell
			 * the face list is made up of polygon entries.  For each entry,
			 *      the length comes first, then a number of data points
			 *      equal to the length.
			 */
			polygon_length = max_info -> face_list[i];
			/* 
			 * hoops syntax: to subtract a polygon from a shell, use negative
			 *      polygon length.  We account for this here with abs. val.
			 */
			if( polygon_length < 0 ) polygon_length = -polygon_length;
			total_info -> face_list[ i + current_face_list_offset ] =  
				max_info -> face_list[i];
			for( j = 1 ; j <= polygon_length ; j++ ) 
			{
				/*
				 * remember, the point list was moved, so the face list 
				 * has to be modified as it is copied
				 */
				total_info -> face_list[ current_face_list_offset + i + j ] = 
					max_info -> face_list[ i + j ] + current_point_list_offset;
			}
			i += polygon_length + 1;
			
		} /* end while ( i < local_face_list_len ) */
		
		HC_Open_Geometry( key );
		for( i = 0 ; i < local_point_list_len ; i++ )
		{
			v = &(total_info -> normal_list[current_point_list_offset+i]);
			HC_Open_Vertex( i );
			HC_Show_Net_Normal( &(v -> x), &(v -> y), &(v -> z) );
			HC_Close_Vertex();
		}
		HC_Close_Geometry();
		
		/* update the current offsets and prepare to load the next shell. */
		total_info -> point_list_len += local_point_list_len;
		total_info -> face_list_len  += local_face_list_len;
		
	} /* end while ( HC_Find_Contents(...) ) */
	HC_End_Contents_Search ();
	
	HC_Flush_Contents (".", "shells");
	
} /* end function Recursive_Merge_Shells */


/*******************************\
 * Merge_Shells
 * Extracts all of the shells from the currently open segment and
 *       its children, and merges them into one.
 * Should there be a boolean argument to indicate whether to search children?
 \*******************************/
HC_KEY Merge_Shells( int do_recurse )
{
	HC_KEY		key;
	Vector	*v;
	int		i;
	Shell_Info   total_info = { 0, NULL, NULL, 0, NULL };
	Shell_Info   max_info   = { 0, NULL, NULL, 0, NULL };
	Shell_Info   optimized  = { 0, NULL, NULL, 0, NULL };
	int		*point_mapping_list, *face_mapping_list;
	
	Segment_Shell_Allocate (&total_info, &max_info, do_recurse);
	Recursive_Merge_Shells (&total_info, &max_info, do_recurse);
	
	optimized.point_list = (Point *) 
		malloc (total_info.point_list_len * sizeof (Point));
	optimized.face_list =  (int *) 
		malloc (total_info.face_list_len * sizeof (int));
	
	point_mapping_list = (int *) 
		malloc (total_info.point_list_len * sizeof (int));
	face_mapping_list = (int *) 
		malloc (total_info.face_list_len * sizeof (int));
	
	/* optimize it */
	HC_Compute_Optimized_Shell ( total_info.point_list_len, total_info.point_list,
							     (float *)total_info.normal_list,
							     total_info.face_list_len, total_info.face_list,
							     "tolerance = 0.2oru, normal tolerance = 30degrees",
							     &(optimized.point_list_len),optimized.point_list,
							     &(optimized.face_list_len), optimized.face_list,
							     point_mapping_list, face_mapping_list );
	key = HC_KInsert_Shell ( optimized.point_list_len, 
							 optimized.point_list,
							 optimized.face_list_len, 
							 optimized.face_list);
	
	HC_Open_Geometry( key );
	for( i = 0 ; i < total_info.point_list_len ; i++ )
	{
		if( point_mapping_list[i] >= 0 )
		{
			v = &(total_info.normal_list[ i ]);
			HC_Open_Vertex( point_mapping_list[i] );
			HC_Set_Normal( v -> x,  v -> y,  v -> z );
			HC_Close_Vertex();
		}
	}
	HC_Close_Geometry();
	
	if(max_info.point_list)		free(max_info.point_list);
	if(max_info.face_list)		free(max_info.face_list);
	if(max_info.face_list)		free(max_info.normal_list);
	if(total_info.point_list)	free(total_info.point_list);
	if(total_info.face_list)	free(total_info.face_list);
	if(total_info.normal_list)	free(total_info.normal_list);
	if(optimized.point_list)	free(optimized.point_list);
	if(optimized.face_list)		free(optimized.face_list);
	if(optimized.normal_list)	free(optimized.normal_list);
	if(point_mapping_list)		free(point_mapping_list);
	if(face_mapping_list)		free(face_mapping_list);

	return key;   
	
} /* end function Merge_Shells */


/* open all of the vertices in a segment, and set the normals to be all the same. */
void set_uniform_normal( HC_KEY key, const Vector *v )
{
	int i, imax, idummy;
	HC_Show_Shell_Size (key,&imax, &idummy);
	HC_Open_Geometry (key);
	for( i = 0 ; i < imax ; i++ ) 
	{
		HC_Open_Vertex( i );
		HC_Set_Normal( v->x, v->y, v->z );
		HC_Close_Vertex();
	}
	HC_Close_Geometry();
	
} /* end function set_uniform_normal */


HC_KEY insert_shells( )
{
	float pointlist_one[9] = { 0,0,0, 1,0,0, 0,1,0 };
	float pointlist_two[9] = { 0,0,0, 0,0,1, 1,0,0 };
	Vector normal_one = { 0,0,-1 };
	Vector normal_two = { 0,0.1f,-1 };
	int connect[4] = { 3, 0, 1, 2 };
	HC_KEY key, tempkey;

	key = HC_KOpen_Segment("");
	tempkey = HC_KInsert_Shell( 3, pointlist_one, 4, connect );
	set_uniform_normal (tempkey, &normal_one);
	tempkey = HC_KInsert_Shell( 3, pointlist_two, 4, connect );
	set_uniform_normal (tempkey, &normal_two);
	HC_Close_Segment();
	return key;
	
} /* end function insert_shells */


void the_usual()
{
	const Point tripod = { 1, 1, 7 };
	const Point origin = { 0,0,0 };
	const Point lpos = { 0,0,7 };
	const Point up = { 0,1,0 };
	const float scenewidth = 3;
	const float sceneheight = 3;
	
	HC_Set_Driver_Options ("subscreen=( -0.3, .3, -0.3, 0.3 )");
	HC_Set_Visibility ("markers = off, edges = off");
	HC_Insert_Distant_Light (lpos.x, lpos.y, lpos.z);
	HC_Set_Camera (&tripod, &origin, &up, scenewidth, sceneheight, "Perspective");
	HC_Set_Color ("blue");
	HC_Set_Selectability ("geometry");
	HC_Set_Rendering_Options ("lighting interpolation = phong");
	
} /* end function the_usual*/


/*******************************\
 * MAIN function
\*******************************/
int main(
		 int	argc,
		 char	**argv )
{
	HC_Open_Segment ("?picture");
	the_usual();
	insert_shells();
	Merge_Shells (ISRECURSIVE);
	
	HC_Update_Display ();
	HC_Write_Metafile (".", "output.hmf", "");
	HC_Close_Segment ();
	
	HC_Pause ();
	HC_Exit_Program ();
	return 0;
	
} /* end function main */


